import { createAssociatedTokenAccountInstruction, createInitializeMintInstruction, createMintToInstruction, createTransferInstruction, getAccount, getAssociatedTokenAddress, getAssociatedTokenAddressSync, TOKEN_PROGRAM_ID, } from '@solana/spl-token';
import { Keypair, PublicKey, SystemProgram, Transaction, } from '@solana/web3.js';
export async function mintTo(provider, mint, amount, destination) {
    const mintTx = new Transaction();
    mintTx.add(createMintToInstruction(mint, destination, provider.wallet.publicKey, amount));
    try {
        await provider.sendAndConfirm(mintTx);
    }
    catch (e) {
        console.log('Error', e, e.logs);
        if (e.logs) {
            console.error(e.logs.join('\n'));
        }
        throw e;
    }
}
export async function createAtaAndTransferInstructions(provider, mint, amount, from = provider.wallet.publicKey, to = provider.wallet.publicKey, payer = provider.wallet.publicKey) {
    const toAta = await getAssociatedTokenAddress(mint, to, true);
    const instructions = [];
    if (!(await provider.connection.getAccountInfo(toAta))) {
        instructions.push(createAssociatedTokenAccountInstruction(payer, toAta, to, mint));
    }
    const fromAta = await getAssociatedTokenAddress(mint, from, true);
    if (amount != 0) {
        instructions.push(createTransferInstruction(fromAta, toAta, from, BigInt(amount.toString())));
    }
    return {
        instructions,
        toAta,
    };
}
export async function createAtaAndTransfer(provider, mint, amount, from = provider.wallet.publicKey, to = provider.wallet.publicKey, authority = provider.wallet.publicKey, payer = provider.wallet.publicKey, confirmOptions) {
    const transferIx = new Transaction();
    const { instructions, toAta } = await createAtaAndTransferInstructions(provider, mint, amount, from, to, payer);
    if (instructions.length > 0)
        transferIx.add(...instructions);
    try {
        if (instructions.length > 0)
            await provider.sendAndConfirm(transferIx, undefined, confirmOptions);
    }
    catch (e) {
        console.log('Error', e, e.logs);
        if (e.logs) {
            console.error(e.logs.join('\n'));
        }
        throw e;
    }
    return toAta;
}
export async function createAtaAndMintInstructions(provider, mint, amount, to = provider.wallet.publicKey, authority = provider.wallet.publicKey, payer = provider.wallet.publicKey) {
    const ata = await getAssociatedTokenAddress(mint, to, true);
    const instructions = [];
    if (!(await provider.connection.getAccountInfo(ata))) {
        instructions.push(createAssociatedTokenAccountInstruction(payer, ata, to, mint));
    }
    if (amount != 0) {
        instructions.push(createMintToInstruction(mint, ata, authority, BigInt(amount.toString())));
    }
    return {
        instructions,
        ata,
    };
}
export async function createAta(provider, mint, owner = provider.wallet.publicKey, authority = provider.wallet.publicKey, payer = provider.wallet.publicKey, confirmOptions) {
    const tx = new Transaction();
    const ata = await getAssociatedTokenAddress(mint, owner, true);
    if (!(await provider.connection.getAccountInfo(ata))) {
        tx.add(createAssociatedTokenAccountInstruction(payer, ata, owner, mint));
    }
    try {
        await provider.sendAndConfirm(tx, undefined, confirmOptions);
    }
    catch (e) {
        console.log('Error', e, e.logs);
        if (e.logs) {
            console.error(e.logs.join('\n'));
        }
        throw e;
    }
    return ata;
}
export async function createAtaAndMint(provider, mint, amount, to = provider.wallet.publicKey, authority = provider.wallet.publicKey, payer = provider.wallet.publicKey, signers = [], confirmOptions) {
    const mintTx = new Transaction();
    const { instructions, ata } = await createAtaAndMintInstructions(provider, mint, amount, to, authority, payer);
    if (instructions.length > 0)
        mintTx.add(...instructions);
    try {
        if (instructions.length > 0)
            await provider.sendAndConfirm(mintTx, signers.length == 0 ? undefined : signers, confirmOptions);
    }
    catch (e) {
        console.log('Error', e, e.logs);
        if (e.logs) {
            console.error(e.logs.join('\n'));
        }
        throw e;
    }
    return ata;
}
export async function createMintInstructions(provider, decimals, mintAuthority, freezeAuthority = null, mintKeypair = Keypair.generate()) {
    const mintKey = mintKeypair.publicKey;
    return [
        SystemProgram.createAccount({
            fromPubkey: provider.wallet.publicKey,
            newAccountPubkey: mintKey,
            space: 82,
            lamports: await provider.connection.getMinimumBalanceForRentExemption(82),
            programId: TOKEN_PROGRAM_ID,
        }),
        await createInitializeMintInstruction(mintKeypair.publicKey, decimals, mintAuthority, freezeAuthority),
    ];
}
export async function createMint(provider, decimals, mintAuthority, freezeAuthority = null, mintKeypair = Keypair.generate()) {
    const tx = new Transaction();
    tx.add(...(await createMintInstructions(provider, decimals, mintAuthority, freezeAuthority, mintKeypair)));
    try {
        await provider.sendAndConfirm(tx, [mintKeypair]);
    }
    catch (e) {
        console.log('Error', e, e.logs);
        if (e.logs) {
            console.error(e.logs.join('\n'));
        }
        throw e;
    }
    return mintKeypair.publicKey;
}
export const isAta = async (connection, address, mint) => {
    let tokenAccount;
    try {
        tokenAccount = await getAccount(connection, address);
    }
    catch (e) {
        return;
    }
    if (mint && !tokenAccount.mint.equals(mint)) {
        return;
    }
    const expectedAta = getAssociatedTokenAddressSync(tokenAccount.mint, tokenAccount.owner, true);
    if (expectedAta.equals(address)) {
        return { owner: tokenAccount.owner };
    }
    return;
};
export const getBalanceDeltas = async (conn, sig, mint = new PublicKey('rndrizKT3MK1iimdxRdWabcF7Zg7AR5T4nud4EkHBof')) => {
    const tx = await conn.getParsedTransaction(sig, {
        maxSupportedTransactionVersion: 0,
        commitment: 'confirmed',
    });
    if (!tx) {
        throw new Error(`tx ${tx} not found`);
    }
    const bals = {};
    for (const b of tx.meta.postTokenBalances) {
        if (b.mint == mint.toString()) {
            bals[b.owner] = Number(b.uiTokenAmount.amount);
        }
    }
    let sender = '';
    for (const b of tx.meta.preTokenBalances) {
        if (b.mint == mint.toString()) {
            bals[b.owner] -= Number(b.uiTokenAmount.amount);
            if (bals[b.owner] < 0) {
                sender = b.owner;
            }
        }
    }
    return {
        sender: new PublicKey(sender),
        deltas: bals,
        executedAt: new Date(tx.blockTime * 1000),
    };
};
/*export async function createNft(
  provider: anchor.AnchorProvider,
  recipient: PublicKey,
  data: any = {},
  collectionKey?: PublicKey,
  mintKeypair: Keypair = Keypair.generate(),
  holderKey: PublicKey = provider.wallet.publicKey
): Promise<{ mintKey: PublicKey; collectionKey: PublicKey | undefined }> {
  const mintKey = mintKeypair.publicKey;

  const instructions = await createMintInstructions(
    provider,
    0,
    provider.wallet.publicKey,
    provider.wallet.publicKey,
    mintKeypair
  );
  const [metadata] = PublicKey.findProgramAddressSync(
    [
      Buffer.from("metadata", "utf8"),
      METADATA_PROGRAM_ID.toBuffer(),
      mintKey.toBuffer(),
    ],
    METADATA_PROGRAM_ID
  );
  instructions.push(
    await createCreateMetadataAccountV3Instruction(
      {
        metadata,
        mint: mintKey,
        mintAuthority: provider.wallet.publicKey,
        payer: provider.wallet.publicKey,
        updateAuthority: provider.wallet.publicKey,
      },
      {
        createMetadataAccountArgsV3: {
          data: {
            name: "test",
            symbol: "TST",
            uri: "https://shdw-drive.genesysgo.net/6tcnBSybPG7piEDShBcrVtYJDPSvGrDbVvXmXKpzBvWP/dc.json",
            sellerFeeBasisPoints: 10,
            creators: [
              {
                address: holderKey,
                verified: true,
                share: 100,
              },
            ],
            collection: collectionKey
              ? { key: collectionKey, verified: false }
              : null,
            uses: null,
            ...data,
          },
          isMutable: true,
          collectionDetails: null,
        },
      }
    )
  );

  const { instructions: mintInstrs } = await createAtaAndMintInstructions(
    provider,
    mintKeypair.publicKey,
    1,
    recipient
  );
  instructions.push(...mintInstrs);

  if (collectionKey) {
    const [collectionMetadataAccount] = PublicKey.findProgramAddressSync(
      [
        Buffer.from("metadata", "utf8"),
        METADATA_PROGRAM_ID.toBuffer(),
        collectionKey.toBuffer(),
      ],
      METADATA_PROGRAM_ID
    );
    const [collectionMasterEdition] = PublicKey.findProgramAddressSync(
      [
        Buffer.from("metadata", "utf8"),
        METADATA_PROGRAM_ID.toBuffer(),
        collectionKey.toBuffer(),
        Buffer.from("edition", "utf8"),
      ],
      METADATA_PROGRAM_ID
    );
    const instruction = createVerifyCollectionInstruction({
      metadata: metadata,
      collectionAuthority: provider.wallet.publicKey,
      payer: provider.wallet.publicKey,
      collectionMint: collectionKey,
      collection: collectionMetadataAccount,
      collectionMasterEditionAccount: collectionMasterEdition,
    });
    instructions.push(instruction);
  }

  const [edition] = PublicKey.findProgramAddressSync(
    [
      Buffer.from("metadata", "utf8"),
      METADATA_PROGRAM_ID.toBuffer(),
      mintKey.toBuffer(),
      Buffer.from("edition", "utf8"),
    ],
    METADATA_PROGRAM_ID
  );
  instructions.push(
    createCreateMasterEditionV3Instruction(
      {
        edition,
        mint: mintKey,
        updateAuthority: provider.wallet.publicKey,
        mintAuthority: provider.wallet.publicKey,
        payer: provider.wallet.publicKey,
        metadata,
      },
      {
        createMasterEditionArgs: {
          maxSupply: 0,
        },
      }
    )
  );

  const tx = new Transaction();
  tx.add(...instructions);

  try {
    await provider.sendAndConfirm(tx, [mintKeypair]);
  } catch (e: any) {
    console.log("Error", e, e.logs);
    if (e.logs) {
      console.error(e.logs.join("\n"));
    }
    throw e;
  }

  return {
    mintKey,
    collectionKey,
  };
}*/
//# sourceMappingURL=token.js.map