import BN from "bn.js";
import { ComputeBudgetProgram, Keypair, PublicKey, Transaction, TransactionInstruction } from "@solana/web3.js";
import { sendAndConfirmWithRetry } from "./transaction";
export const truthy = (value) => !!value;
export function toNumber(numberOrBn, mintOrDecimals) {
    if (BN.isBN(numberOrBn) || typeof numberOrBn == "bigint") {
        return amountAsNum(numberOrBn, mintOrDecimals);
    }
    else {
        return numberOrBn;
    }
}
export function amountAsNum(amount, mintOrDecimals) {
    const decimals = typeof mintOrDecimals === "number"
        ? mintOrDecimals
        : mintOrDecimals.decimals;
    const factor = new BN(Math.pow(10, decimals).toString());
    const decimal = new BN(amount.toString()).mod(factor).toNumber() / factor.toNumber();
    return new BN(amount.toString()).div(factor).toNumber() + decimal;
}
export function toBN(numberOrBn, mintOrDecimals) {
    const decimals = typeof mintOrDecimals === "number"
        ? mintOrDecimals
        : mintOrDecimals.decimals;
    if (BN.isBN(numberOrBn)) {
        return numberOrBn;
    }
    else {
        return new BN(Number(Math.floor(Number(numberOrBn))))
            .mul(new BN(Math.pow(10, decimals)))
            .add(new BN(Math.round((Number(numberOrBn) % 1) * Math.pow(10, decimals))));
    }
}
export function supplyAsNum(mint) {
    return amountAsNum(new BN(mint.supply.toString()), mint);
}
export function numberWithCommas(x, decimals = 4) {
    return roundToDecimals(x, decimals).toLocaleString("en-US", {
        maximumFractionDigits: decimals,
    });
}
export function roundToDecimals(num, decimals) {
    return Math.trunc(num * Math.pow(10, decimals)) / Math.pow(10, decimals);
}
export function humanReadableBigint(bigint, decimals, round = decimals) {
    return numberWithCommas(roundToDecimals(toNumber(bigint, decimals), round));
}
export function humanReadable(bn, decimalsOrMint) {
    return numberWithCommas(roundToDecimals(toNumber(bn, decimalsOrMint), typeof decimalsOrMint == "number"
        ? decimalsOrMint
        : decimalsOrMint.decimals));
}
export const getSolanaKeypair = (secretKey) => Keypair.fromSecretKey(Buffer.from(secretKey, "base64"));
export function* packInstructions(as, prioFee = 60000, computeLimit = 0, additionalIxs = []) {
    let newTx = (startIxs) => {
        let tx = new Transaction({
            recentBlockhash: PublicKey.default.toString(),
            feePayer: PublicKey.default
        });
        if (prioFee) {
            tx.add(ComputeBudgetProgram.setComputeUnitPrice({
                microLamports: prioFee,
            }));
        }
        if (computeLimit) {
            tx.add(ComputeBudgetProgram.setComputeUnitLimit({ units: computeLimit }));
        }
        if (additionalIxs.length > 0) {
            tx.add(...additionalIxs);
        }
        if (startIxs.length > 0) {
            tx.add(...startIxs);
        }
        return tx;
    };
    let tx = newTx([]);
    let txAs = [];
    for (const a of as) {
        tx.add(a.ix);
        try {
            const bl = [...tx.serialize({
                    requireAllSignatures: false,
                    verifySignatures: false
                })].length;
            if (bl > 1232) {
                console.log(`transaction size limit reached: ${bl}. aborting`);
                yield txAs;
                tx = newTx([a.ix]);
                txAs = [a];
            }
            txAs.push(a);
        }
        catch (e) {
            yield txAs;
            tx = newTx([a.ix]);
            txAs = [a];
        }
    }
    yield txAs;
}
export const pingMillis = async (connection, signer) => {
    const now = Date.now();
    const tx = new Transaction({
        feePayer: signer.publicKey,
        recentBlockhash: (await connection.getLatestBlockhash()).blockhash,
    }).add(new TransactionInstruction({
        keys: [{ pubkey: signer.publicKey, isSigner: true, isWritable: true }],
        data: Buffer.from("ping", "utf-8"),
        programId: new PublicKey("MemoSq4gqABAXKb96qnH8TysNcWxMyWCqXgDLGmfcHr"),
    }));
    tx.sign(signer);
    await sendAndConfirmWithRetry(connection, tx.serialize(), { skipPreflight: true }, 'confirmed', 300000);
    return Date.now() - now;
};
//# sourceMappingURL=utils.js.map